# cgFireInTheSky.tcl
#
# Original files from: http://www.shadertech.com/shaders/FireInTheSky-src.zip
#
# Original files are Copyright (c) 2002 Jason Jerald
# This software is provided 'as-is', without any express or implied warranty.
# In no event will the authors be held liable for any damages arising from the
# use of this software.
# 
# Permission is granted to anyone to use this software for any purpose, 
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
# 
# 1. The origin of this software must not be misrepresented;
# you must not claim that you wrote the original software. 
# If you use this software in a product, an acknowledgment in 
# the product documentation would be appreciated but is not required.
# 
# 2. Altered source versions must be plainly marked as such,
# and must not be misrepresented as being the original software.
# 
# 3. This notice may not be removed or altered from any source distribution.
#
# Modified for Tcl3D by Paul Obermeier 2005/11/07
# See www.tcl3d.org for the Tcl3D extension.

package require Tk
package require tcl3d 0.3.2

if { [info procs tcl3dHaveCg] ne "" } {
    if { ![tcl3dHaveCg] } {
        tk_messageBox -icon error -type ok -title "Error" \
                      -message "You do not have Cg installed."
        exit
    }
}

set TEXTUREWIDTH 64
#set TEXTUREWIDTH 512

set NoiseTextureMap [tcl3dVector GLfloat [expr $::TEXTUREWIDTH*$::TEXTUREWIDTH]]

set rRotationX 0.0
set rRotationY 0.0

# Show errors occuring in the Togl callbacks.
proc bgerror { msg } {
    tk_messageBox -icon error -type ok -message "Error: $msg"
    exit
}

# Print info message into widget a the bottom of the window.
proc PrintInfo { msg } {
    if { [winfo exists .fr.info] } {
        .fr.info configure -text $msg
    }
}

proc GetFileName { fileName } {
    return [tcl3dGetExtFile [file join [file dirname [info script]] $fileName]]
}

proc InitCgShader {} {
    set ::lightPos { 0.0 1.0 0.0 }

    set ::profile [tcl3dCgFindProfile $::CG_PROFILE_ARBFP1 $::CG_PROFILE_FP30]
    if { $::profile eq "" } {
        error "Could not find Cg fragment profile ARBFP1 or FP30"
        exit
    }

    # Create cgContext.
    set ::context [cgCreateContext]

    # Add shaders into context
    set ::CloudShader [cgCreateProgramFromFile $::context \
                       CG_SOURCE [GetFileName "CloudShader.cg"] \
                       $::profile "main" "NULL"]
    set retVal [tcl3dCgGetError $::context]
    if { $retVal ne "" } {
        error "cgCreateProgramFromFile CloudShader: $retVal"
    }
    tcl3dCgPrintProgramInfo $::CloudShader "CloudShader.cg"

    set ::OceanShader [cgCreateProgramFromFile $::context \
                       CG_SOURCE [GetFileName "OceanShader.cg"] \
                       $::profile "main2" "NULL"]
    set retVal [tcl3dCgGetError $::context]
    if { $retVal ne "" } {
        error "cgCreateProgramFromFile OceanShader: $retVal"
    }
    tcl3dCgPrintProgramInfo $::OceanShader "OceanShader.cg"

    set ::SunShader [cgCreateProgramFromFile $::context \
                     CG_SOURCE [GetFileName "SunShader.cg"] \
                     $::profile "main3" "NULL"]
    set retVal [tcl3dCgGetError $::context]
    if { $retVal ne "" } {
        error "cgCreateProgramFromFile SunShader: $retVal"
    }
    tcl3dCgPrintProgramInfo $::SunShader "SunShader.cg"

    # noise
    set ::NoiseTextureParam [cgGetNamedParameter $::CloudShader "Noise"]

    # light position
    set ::LightPosParam [cgGetNamedParameter $::OceanShader "LightPos"]

    # Load our programs
    cgGLLoadProgram $::CloudShader
    set retVal [tcl3dCgGetError $::context]
    if { $retVal ne "" } {
        error "cgGLLoadProgram CloudShader: $retVal"
    }

    cgGLLoadProgram $::OceanShader
    set retVal [tcl3dCgGetError $::context]
    if { $retVal ne "" } {
        error "cgGLLoadProgram OceanShader: $retVal"
    }

    cgGLLoadProgram $::SunShader
    set retVal [tcl3dCgGetError $::context]
    if { $retVal ne "" } {
        error "cgGLLoadProgram SunShader: $retVal"
    }

    CreateNoiseTexture

    # Create texture surface
    set ::texName [tcl3dVector GLuint 1]
    glGenTextures 1 $::texName
    glBindTexture GL_TEXTURE_2D [$::texName get 0]

    glEnable GL_TEXTURE_2D
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MAG_FILTER $::GL_LINEAR
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MIN_FILTER $::GL_LINEAR
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_WRAP_S $::GL_REPEAT
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_WRAP_T $::GL_REPEAT
    glTexImage2D GL_TEXTURE_2D 0 $::GL_LUMINANCE \
                 $::TEXTUREWIDTH $::TEXTUREWIDTH \
                 0 GL_LUMINANCE GL_FLOAT $::NoiseTextureMap
    glDisable GL_TEXTURE_2D
}

proc Render {} {
    glClear GL_COLOR_BUFFER_BIT
    glColor3f 0.0 0.0 0.0
    glEnable GL_TEXTURE_2D
    glBindTexture GL_TEXTURE_2D [$::texName get 0]

    # the sun
    cgGLBindProgram $::SunShader
    cgGLEnableProfile $::profile
    glTranslatef 0.0 -0.1 -10.0
    glutSolidSphere 1.5 30 30
    glTranslatef -1.0 0.1 10.0

    glTranslatef 0.0 -0.2 0.0

    # the sky
    glEnable GL_BLEND
    glBlendFunc  GL_ONE GL_ONE_MINUS_SRC_ALPHA

    cgGLBindProgram $::CloudShader

    # Prepare first texture
    # OPA TODO cgGLActiveTexture $::NoiseTextureMap

    # the light position

    cgGLSetParameter4f $::LightPosParam \
                       [lindex $::lightPos 0] [lindex $::lightPos 1] \
                       [lindex $::lightPos 2] 1.0

    glBegin GL_QUADS
        glTexCoord2f 0.0 0.0; glVertex3f -10.0 1.0   1.0
        glTexCoord2f 1.0 0.0; glVertex3f  10.0 1.0   1.0
        glTexCoord2f 1.0 1.0; glVertex3f  10.0 0.0 -10.0
        glTexCoord2f 0.0 1.0; glVertex3f -10.0 0.0 -10.0
    glEnd
    glDisable GL_BLEND

    # the ocean
    cgGLBindProgram $::OceanShader
    glBegin GL_QUADS
        glTexCoord2f 0 0; glVertex3f -10.0 -1.0   1.0
        glTexCoord2f 1 0; glVertex3f  10.0 -1.0   1.0
        glTexCoord2f 1 1; glVertex3f  10.0  0.0 -10.0
        glTexCoord2f 0 1; glVertex3f -10.0  0.0 -10.0
    glEnd

    cgGLDisableProfile $::profile
}

proc CreateNoiseTexture {} {
    for { set i 0 } { $i < $::TEXTUREWIDTH } { incr i } {
        for { set j 0 } { $j < $::TEXTUREWIDTH } { incr j } {
            set ind [expr $i*$::TEXTUREWIDTH + $j]
            $::NoiseTextureMap set $ind [expr ([PRNGenerator $ind] + 1.0) / 2]
        }
    }
}

proc PRNGenerator { x } {
    set x [expr ($x << 13) ^ $x]

    set Prime1 15731
    set Prime2 789221
    set Prime3 1376312589

    return [expr (1.0 - (($x * ($x*$x*$Prime1 + $Prime2) + $Prime3) & \
                  0x7fffffff) / 1073741824.0)]
}

proc tclDisplayFunc { toglwin } {
    glClear [expr $::GL_COLOR_BUFFER_BIT | $::GL_DEPTH_BUFFER_BIT]

    glMatrixMode GL_MODELVIEW
    glPushMatrix
    glLoadIdentity
    gluLookAt 0 0 1 0 0 0 0 1 0
    glRotatef $::rRotationX 0 1 0
    glRotatef $::rRotationY 1 0 0

    Render

    glPopMatrix
    $toglwin swapbuffers
}

proc tclCreateFunc { toglwin } {
    glClearColor 0.0 0.0 0.0 1.0
    glDisable GL_LIGHTING

    InitCgShader
}

proc tclReshapeFunc { toglwin w h } {
    if { $h == 0 } {
        set height 1
    }
  
    glViewport 0 0 $w $h
    
    glMatrixMode GL_PROJECTION
    glLoadIdentity
  
    # Calculate the aspect ratio of the window */
    gluPerspective 60.0 [expr double($w)/double($h)] 0.01 1000.0

    glMatrixMode GL_MODELVIEW
    glLoadIdentity
}

proc CreateWindow {} { 
    frame .fr
    pack .fr -expand 1 -fill both
    togl .fr.toglwin -width 512 -height 512 \
                     -double true -depth true -alpha true \
                     -createproc  tclCreateFunc \
                     -reshapeproc tclReshapeFunc \
                     -displayproc tclDisplayFunc
    label .fr.info

    grid .fr.toglwin -row 0 -column 0 -sticky news
    grid .fr.info    -row 1 -column 0 -sticky news
    grid rowconfigure .fr 0 -weight 1
    grid columnconfigure .fr 0 -weight 1

    wm title . "Tcl3D demo: Fire in the sky"

    bind . <Key-Escape> "exit"
}

CreateWindow

PrintInfo [format "Running on %s with a %s (OpenGL %s, Tcl %s)" \
           $tcl_platform(os) [glGetString GL_RENDERER] \
           [glGetString GL_VERSION] [info patchlevel]]
