# checkerBoard.tcl
#
# This program creates a checkerboard image in two ways.
# The first texture is created with an algorithm, as used in some of the
# RedBook examples (ex. checker.tcl). This algorithm has been converted 1:1
# from C to Tcl. Very slow.
# The second image is created using the Img extension, which is essentially 
# faster.
#
# Author: Paul Obermeier
# Date: 2006-09-22

package require tcl3d 0.3
package require Img

# Create checkerboard texture
set imgWidth  64
set imgHeight 64

# Font to be used in the Tk listbox.
set listFont {-family {Courier} -size 10}

# Show errors occuring in the Togl callbacks.
proc bgerror { msg } {
    tk_messageBox -icon error -type ok -message "Error: $msg"
    exit
}

# Print info message into widget a the bottom of the window.
proc PrintInfo { msg } {
    if { [winfo exists .fr.info] } {
        .fr.info configure -text $msg
    }
}

# The classical method converted 1:1 from C to Tcl. Very slow.
proc makeCheckImage1 {} {
    set ::checkImage1 [tcl3dVector GLubyte [expr $::imgHeight*$::imgWidth*4]]

    for { set i 0 } { $i < $::imgHeight } { incr i } {
        for { set j 0 } { $j < $::imgWidth } { incr j } { 
         set c [expr {(((($i&0x8)==0)^(($j&0x8))==0))*255}]
         $::checkImage1 set [expr {($i*$::imgWidth + $j)*4 + 0}] $c
         $::checkImage1 set [expr {($i*$::imgWidth + $j)*4 + 1}] $c
         $::checkImage1 set [expr {($i*$::imgWidth + $j)*4 + 2}] $c
         $::checkImage1 set [expr {($i*$::imgWidth + $j)*4 + 3}] 255
      }
   }
}

# Create the checker image with the use of the Img extension. Fast.
proc makeCheckImage2 {} {
    set tmp [image create photo -width 2 -height 2]
    set img [image create photo -width $::imgWidth -height $::imgHeight]
    $tmp put {{#ffffff #FF0000} {#FF0000 #ffffff}} 
    $img copy $tmp -from 0 0 2 2 -zoom 8 8 -to 0 0 $::imgWidth $::imgHeight
    set ::checkImage2 [tcl3dVectorFromPhoto $img]
    image delete $tmp
    image delete $img
}

proc tclCreateFunc { toglwin } {    
    glClearColor 0.5 0.5 0.5 0.0
    glShadeModel GL_FLAT
    glEnable GL_DEPTH_TEST
}

proc CreateCheckerImgs {} {
    set t1 [time makeCheckImage1]
    set t2 [time makeCheckImage2]
    .fr.usage insert end "Black checker: $t1"
    .fr.usage insert end "Red   checker: $t2"

    glPixelStorei GL_UNPACK_ALIGNMENT 1

    set ::texName [tcl3dVector GLuint 2]
    glGenTextures 2 $::texName

    glBindTexture GL_TEXTURE_2D [$::texName get 0]
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_WRAP_S $::GL_REPEAT
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_WRAP_T $::GL_REPEAT
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MAG_FILTER $::GL_NEAREST
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MIN_FILTER $::GL_NEAREST

    glTexImage2D GL_TEXTURE_2D 0 $::GL_RGBA \
                 $::imgWidth $::imgHeight \
                 0 GL_RGBA GL_UNSIGNED_BYTE $::checkImage1

    glBindTexture GL_TEXTURE_2D [$::texName get 1]
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_WRAP_S $::GL_REPEAT
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_WRAP_T $::GL_REPEAT
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MAG_FILTER $::GL_NEAREST
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MIN_FILTER $::GL_NEAREST

    glTexImage2D GL_TEXTURE_2D 0 $::GL_RGBA \
                 $::imgWidth $::imgHeight \
                 0 GL_RGBA GL_UNSIGNED_BYTE $::checkImage2
}

proc tclDisplayFunc { toglwin } {
    glClear [expr $::GL_COLOR_BUFFER_BIT | $::GL_DEPTH_BUFFER_BIT]
    glEnable GL_TEXTURE_2D
    glTexEnvf GL_TEXTURE_ENV GL_TEXTURE_ENV_MODE $::GL_DECAL

    glBindTexture GL_TEXTURE_2D [$::texName get 0]
    glBegin GL_QUADS
        glTexCoord2f 0.0 0.0 ; glVertex3f -2.0 -1.0 0.0
        glTexCoord2f 0.0 1.0 ; glVertex3f -2.0 1.0 0.0
        glTexCoord2f 1.0 1.0 ; glVertex3f 0.0 1.0 0.0
        glTexCoord2f 1.0 0.0 ; glVertex3f 0.0 -1.0 0.0
    glEnd

    glBindTexture GL_TEXTURE_2D [$::texName get 1]
    glBegin GL_QUADS
        glTexCoord2f 0.0 0.0 ; glVertex3f 0.0 -1.0 0.0
        glTexCoord2f 0.0 1.0 ; glVertex3f 0.0 1.0 0.0
        glTexCoord2f 1.0 1.0 ; glVertex3f 2.0 1.0 0.0
        glTexCoord2f 1.0 0.0 ; glVertex3f 2.0 -1.0 0.0
    glEnd
    glFlush
    glDisable GL_TEXTURE_2D
}

proc tclReshapeFunc { toglwin w h } {
    glViewport 0 0 $w $h
    glMatrixMode GL_PROJECTION
    glLoadIdentity
    gluPerspective 60.0 [expr double($w)/double($h)] 1.0 30.0
    glMatrixMode GL_MODELVIEW
    glLoadIdentity
    glTranslatef 0.0 0.0 -3.6
}

frame .fr
pack .fr -expand 1 -fill both
togl .fr.toglwin -width 250 -height 250 -double false -depth true \
                 -createproc tclCreateFunc \
                 -reshapeproc tclReshapeFunc \
                 -displayproc tclDisplayFunc 
listbox .fr.usage -font $::listFont -height 3
label   .fr.info
grid .fr.toglwin -row 0 -column 0 -sticky news
grid .fr.usage   -row 1 -column 0 -sticky news
grid .fr.info    -row 2 -column 0 -sticky news
grid rowconfigure .fr 0 -weight 1
grid columnconfigure .fr 0 -weight 1
wm title . "Tcl3D demo: Texture generation comparison"

bind . <Key-Escape> "exit"

.fr.usage insert end "Key-Escape Exit"

CreateCheckerImgs
.fr.usage configure -state disabled

PrintInfo [format "Running on %s with a %s (OpenGL %s, Tcl %s)" \
           $tcl_platform(os) [glGetString GL_RENDERER] \
           [glGetString GL_VERSION] [info patchlevel]]
