# ogl_fps_controls.tcl
#
# This sample demonstrates how to collect user input and 
# build a custom view matrix for First Person Shooter style controls.
#
# Original C++ code by Kevin Harris (kevin@codesampler.com) 02/01/05 
# See www.codesampler.com for the original files
# OpenGL samples page 5: First Person Shooter Controls
#
# Modified for Tcl3D by Paul Obermeier 2005/11/05
# See www.tcl3d.org for the Tcl3D extension.

package require Tk
package require tcl3d 0.2

# Font to be used in the Tk listbox.
set listFont {-family {Courier} -size 10}

set g_WinWidth  640
set g_WinHeight 480

set g_fMoveSpeed 25.0

set g_vEye   [tcl3dVectorFromArgs GLfloat 5.0 5.0 5.0]     ; # Eye Position
set g_vLook  [tcl3dVectorFromArgs GLfloat -0.5 -0.5 -0.5]  ; # Look Vector
set g_vUp    [tcl3dVectorFromArgs GLfloat 0.0 1.0 0.0]     ; # Up Vector
set g_vRight [tcl3dVectorFromArgs GLfloat 1.0 0.0 0.0]     ; # Right Vector

set g_StopWatch [tcl3dNewSwatch]
tcl3dStartSwatch $g_StopWatch
set ::g_LastMousePosX(1) 0
set ::g_LastMousePosY(1) 0

set tmpMat [tcl3dVector GLfloat 16]
set tmpVec [tcl3dVector GLfloat 3]

# red   = +x Axis
# green = +y Axis
# blue  = +z Axis
set vtxArray [tcl3dVectorFromArgs GLfloat \
        0.0 0.0 0.0 \
        5.0 0.0 0.0 \
        0.0 0.0 0.0 \
        0.0 5.0 0.0 \
        0.0 0.0 0.0 \
        0.0 0.0 5.0]

set colorArray [tcl3dVectorFromArgs GLubyte \
        255   0   0 255 \
        255   0   0 255 \
          0 255   0 255 \
          0 255   0 255 \
          0   0 255 255 \
          0   0 255 255]

# Show errors occuring in the Togl callbacks.
proc bgerror { msg } {
    tk_messageBox -icon error -type ok -message "Error: $msg"
    exit
}

# Print info message into widget a the bottom of the window.
proc PrintInfo { msg } {
    if { [winfo exists .fr.info] } {
        .fr.info configure -text $msg
    }
}

proc GetMouseInput { x y } {
    set nXDiff [expr ($x - $::g_LastMousePosX(1))]
    set nYDiff [expr ($y - $::g_LastMousePosY(1))]
        
    if { $nYDiff != 0 } {
        tcl3dMatfRotate [expr -1.0*$nYDiff / 3.0] $::g_vRight $::tmpMat
        tcl3dMatfTransformVector $::g_vLook $::tmpMat $::g_vLook
        tcl3dMatfTransformVector $::g_vUp $::tmpMat $::g_vUp
    }

    if { $nXDiff != 0 } {
        set vec [tcl3dVectorFromArgs GLfloat 0 1 0]
        tcl3dMatfRotate [expr -1.0*$nXDiff / 3.0] $vec $::tmpMat
        tcl3dMatfTransformVector $::g_vLook $::tmpMat $::g_vLook
        tcl3dMatfTransformVector $::g_vUp $::tmpMat $::g_vUp
        $vec delete
    }

    set ::g_LastMousePosX(1) $x
    set ::g_LastMousePosY(1) $y
    .fr.toglwin postredisplay
}

proc MoveForward {} {
    tcl3dVec3fScale [expr {-1.0 * $::g_fMoveSpeed * $::g_fElapsedTime}] $::g_vLook $::tmpVec
    tcl3dVec3fSubtract $::g_vEye $::tmpVec $::g_vEye
    .fr.toglwin postredisplay
}

proc MoveBackward {} {
    tcl3dVec3fScale [expr {-1.0 * $::g_fMoveSpeed * $::g_fElapsedTime}] $::g_vLook $::tmpVec
    tcl3dVec3fAdd $::g_vEye $::tmpVec $::g_vEye
    .fr.toglwin postredisplay
}

proc MoveLeft {} {
    tcl3dVec3fScale [expr {$::g_fMoveSpeed * $::g_fElapsedTime}] $::g_vRight $::tmpVec
    tcl3dVec3fSubtract $::g_vEye $::tmpVec $::g_vEye
    .fr.toglwin postredisplay
}

proc MoveRight {} {
    tcl3dVec3fScale [expr {$::g_fMoveSpeed * $::g_fElapsedTime}] $::g_vRight $::tmpVec
    tcl3dVec3fAdd $::g_vEye $::tmpVec $::g_vEye
    .fr.toglwin postredisplay
}

proc MoveUp {} {
    $::tmpVec set 0 0
    $::tmpVec set 1 [expr {$::g_fMoveSpeed * $::g_fElapsedTime}]
    $::tmpVec set 2 0
    tcl3dVec3fAdd $::g_vEye $::tmpVec $::g_vEye
    .fr.toglwin postredisplay
}

proc MoveDown {} {
    $::tmpVec set 0 0
    $::tmpVec set 1 [expr {$::g_fMoveSpeed * $::g_fElapsedTime}]
    $::tmpVec set 2 0
    tcl3dVec3fSubtract $::g_vEye $::tmpVec $::g_vEye
    .fr.toglwin postredisplay
}

proc tclCreateFunc { toglwin } {
    glClearColor 0.35 0.53 0.7 1.0

    glMatrixMode GL_PROJECTION
    glLoadIdentity
    gluPerspective 45.0 [expr double($::g_WinWidth) / double($::g_WinHeight)] 0.1 100.0
}

proc tclReshapeFunc { toglwin w h } {
    glViewport 0 0 $w $h
    glMatrixMode GL_PROJECTION
    glLoadIdentity
    gluPerspective 45.0 [expr double($w)/double($h)] 0.1 100.0
}

#-----------------------------------------------------------------------------
# Name : updateViewMatrix() 
# Desc : Builds a view matrix suitable for OpenGL.
#
# Here's what the final view matrix should look like:
#
#  |  rx   ry   rz  -(r.e) |
#  |  ux   uy   uz  -(u.e) |
#  | -lx  -ly  -lz   (l.e) |
#  |   0    0    0     1   |
#
# Where r = Right vector
#       u = Up vector
#       l = Look vector
#       e = Eye position in world space
#       . = Dot-product operation
#
#-----------------------------------------------------------------------------

proc updateViewMatrix {} {
    tcl3dMatfIdentity $::tmpMat

    tcl3dVec3fNormalize $::g_vLook

    tcl3dVec3fCrossProduct $::g_vLook $::g_vUp $::g_vRight
    tcl3dVec3fNormalize $::g_vRight

    tcl3dVec3fCrossProduct $::g_vRight $::g_vLook $::g_vUp
    tcl3dVec3fNormalize $::g_vUp

    set rightX [$::g_vRight get 0]
    set rightY [$::g_vRight get 1]
    set rightZ [$::g_vRight get 2]
    set upX    [$::g_vUp get 0]
    set upY    [$::g_vUp get 1]
    set upZ    [$::g_vUp get 2]
    set lookX  [$::g_vLook get 0]
    set lookY  [$::g_vLook get 1]
    set lookZ  [$::g_vLook get 2]

    $::tmpMat set  0 $rightX
    $::tmpMat set  1 $upX
    $::tmpMat set  2 [expr -1.0 * $lookX]
    $::tmpMat set  3 0

    $::tmpMat set  4 $rightY
    $::tmpMat set  5 $upY
    $::tmpMat set  6 [expr -1.0 * $lookY]
    $::tmpMat set  7 0

    $::tmpMat set  8 $rightZ
    $::tmpMat set  9 $upZ
    $::tmpMat set 10 [expr -1.0 * $lookZ]
    $::tmpMat set 11 0

    $::tmpMat set 12 [expr -1.0 * [tcl3dVec3fDotProduct $::g_vRight $::g_vEye]]
    $::tmpMat set 13 [expr -1.0 * [tcl3dVec3fDotProduct $::g_vUp    $::g_vEye]]
    $::tmpMat set 14 [expr  1.0 * [tcl3dVec3fDotProduct $::g_vLook  $::g_vEye]]
    $::tmpMat set 15 1

    set viewList [tcl3dVectorToList $::tmpMat 16]
    glMultMatrixf $viewList
}

proc tclDisplayFunc { toglwin } {
    set startTime [tcl3dLookupSwatch $::g_StopWatch]

    glClear [expr $::GL_COLOR_BUFFER_BIT | $::GL_DEPTH_BUFFER_BIT]

    glMatrixMode GL_MODELVIEW
    glLoadIdentity

    updateViewMatrix

    glColor3f 1.0 1.0 1.0
    glutWireTeapot 2.0

    glEnableClientState GL_VERTEX_ARRAY
    glEnableClientState GL_COLOR_ARRAY
 
    glVertexPointer 3 GL_FLOAT 0 $::vtxArray
    glColorPointer  4 GL_UNSIGNED_BYTE 0 $::colorArray
    glDrawArrays GL_LINES 0 6

    $toglwin swapbuffers
    set endTime [tcl3dLookupSwatch $::g_StopWatch]
    set ::g_fElapsedTime [expr {$endTime - $startTime}]
}

frame .fr
pack .fr -expand 1 -fill both
togl .fr.toglwin -width $g_WinWidth -height $g_WinHeight -double true \
                 -createproc tclCreateFunc \
                 -reshapeproc tclReshapeFunc \
                 -displayproc tclDisplayFunc 
listbox .fr.usage -font $::listFont -height 8
label   .fr.info
grid .fr.toglwin -row 0 -column 0 -sticky news
grid .fr.usage   -row 1 -column 0 -sticky news
grid .fr.info    -row 2 -column 0 -sticky news
grid rowconfigure .fr 0 -weight 1
grid columnconfigure .fr 0 -weight 1
wm title . "Tcl3D demo: First person shooter controls"

bind . <Key-Escape> "exit"
bind . <Key-Up> "MoveForward"
bind . <Key-Down> "MoveBackward"
bind . <Key-Left> "MoveLeft"
bind . <Key-Right> "MoveRight"
bind . <Key-Home> "MoveUp"
bind . <Key-End> "MoveDown"
bind .fr.toglwin <1> {set ::g_LastMousePosX(1) %x; set ::g_LastMousePosY(1) %y}
bind .fr.toglwin <B1-Motion> "GetMouseInput %x %y"

.fr.usage insert end "Mouse-1    Move view with mouse"
.fr.usage insert end "Key-Up     View moves forward"
.fr.usage insert end "Key-Down   View moves backward"
.fr.usage insert end "Key-Left   View strafes to the left"
.fr.usage insert end "Key-Right  View strafes to the right"
.fr.usage insert end "Key-Home   View elevates up"
.fr.usage insert end "Key-End    View elevates down"
.fr.usage insert end "Key-Escape Exit"
.fr.usage configure -state disabled

PrintInfo [format "Running on %s with a %s (OpenGL %s, Tcl %s)" \
           $tcl_platform(os) [glGetString GL_RENDERER] \
           [glGetString GL_VERSION] [info patchlevel]]
